import torch
import torch.nn as nn

__all__ = [
    'ConcatEmbeddings',
    'PassThrough',
    'MeanOfEmbeddings',
]


class ConcatEmbeddings(nn.Module):
    def __init__(self, fields):
        super().__init__()
        self.output_dim = sum([field.output_dim for field in fields.values()])
        self.embedders = nn.ModuleList([field.build_embedder() for field in fields.values()])

    def forward(self, x):
        res = [embedder(values) for embedder, values in zip(self.embedders, x)]
        return torch.cat(res, dim=1)


class PassThrough(nn.Module):
    def forward(self, x):
        return x


class MeanOfEmbeddings(nn.Module):
    def __init__(self, vocab_size, emb_dim):
        super().__init__()
        self.emb = nn.Embedding(vocab_size, emb_dim, padding_idx=0)

    def forward(self, x):
        mask = (x != 0).float()[:, :, None]
        emb = self.emb(x) * mask.float()
        s = mask.squeeze(2).sum(1).clamp_min(1.)[:, None].float()
        return emb.sum(dim=1) / s



import os
import torch
from torch.utils.data import Dataset, DataLoader
import lmdb
import tempfile
import msgpack
import struct

__all__ = [
    'OnFireDataLoader',
    'OnFireDataset',
]


class OnFireDataLoader(DataLoader):
    def __init__(self, data, tfms, batch_size, shuffle=False, num_workers=0,
                 sampler=None, pin_memory=None, drop_last=False, **kwargs):
        num_workers = num_workers if num_workers else os.cpu_count()
        pin_memory = pin_memory if pin_memory != None else torch.cuda.is_available()
        self.ds = OnFireDataset(data, max_readers=num_workers)
        self.tfms = tfms if isinstance(tfms, (list, tuple)) else [tfms]
        super().__init__(self.ds, batch_size=batch_size, shuffle=shuffle,
                         num_workers=num_workers, collate_fn=self.__collate, sampler=sampler,
                         pin_memory=pin_memory, drop_last=drop_last, **kwargs)

    def __collate(self, batch):
        return tuple([tfm(batch) for tfm in self.tfms])


class OnFireDataset(Dataset):
    def __init__(self, data, max_readers):
        self.use_lmdb = max_readers > 1
        if self.use_lmdb:
            tmpdir = tempfile.TemporaryDirectory().name
            self.db = lmdb.open(tmpdir, map_size=1024**4, lock=False, max_readers=max_readers)
            self.key_struct = struct.Struct("!q")
            it = [(self.key_struct.pack(i), msgpack.packb(x)) for i, x in enumerate(data)]
            with self.db.begin(write=True) as txn:
                with txn.cursor() as cursor:
                    cursor.putmulti(it)
        else:
            self.data = data
        self._len = len(data)

    def __getitem__(self, idx):
        if self.use_lmdb:
            key = self.key_struct.pack(idx)
            with self.db.begin() as txn:
                return msgpack.unpackb(txn.get(key))
        else:
            return self.data[idx]

    def __len__(self):
        return self._len



import torch
import numpy as np
from collections import Counter
from sklearn.base import TransformerMixin, BaseEstimator
from unidecode import unidecode

__all__ = [
    'Projector',
    'LabelEncoder',
    'BasicTokenizer',
    'TokensEncoder',
    'ToTensor',
    'MultiLabelEncoder',
    'To2DFloatArray',
    'Log',
]


class Projector(TransformerMixin, BaseEstimator):
    def __init__(self, keys):
        self.keys = keys if isinstance(keys, list) else [keys]

    def fit(self, X, y=None):
        return self

    def _get(self, x):
        for key in self.keys:
            x = x.get(key)
        return x

    def transform(self, X):
        return [self._get(x) for x in X]

    def _inverse(self, x):
        t = x
        for key in reversed(self.keys):
            t = {key: t}
        return t

    def inverse_transform(self, X):
        return [self._inverse(x) for x in X]


class LabelEncoder(TransformerMixin, BaseEstimator):
    class UnknownLabel:
        def __repr__(self):
            return "<UNK>"

    def __init__(self, is_target=False):
        self.is_target = is_target

    def fit(self, X, y=None):
        self.vocab = sorted([x for x in set(X) if x is not None])
        if not self.is_target:
            self.vocab.insert(0, self.UnknownLabel())
        self.category2code = {x: i for i, x in enumerate(self.vocab)}
        return self

    def _get_category_code(self, x):
        return self.category2code.get(x) if self.is_target else self.category2code.get(x, 0)

    def transform(self, X):
        return np.array([self._get_category_code(x) for x in X], dtype=np.int)

    def inverse_transform(self, X):
        return [self.vocab[x] for x in X]


class BasicTokenizer(TransformerMixin, BaseEstimator):
    def __init__(self, lower=True, map_to_ascii=True):
        self.lower = lower
        self.map_to_ascii = map_to_ascii

    def _preprocess(self, text):
        if text is None:
            text = ""
        elif not isinstance(text, str):
            text = str(text)

        if self.map_to_ascii:
            text = unidecode(text)
        if self.lower:
            text = text.lower()
        return text

    def _tokenize(self, text):
        res = []
        for token in text.split():
            while token and not token[-1].isalnum():
                token = token[:-1]
            while token and not token[0].isalnum():
                token = token[1:]
            if token:
                res.append(token)
        return res

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = [self._preprocess(x) for x in X]
        transformed = {hash(x): self._tokenize(x) for x in set(X)}
        return [transformed[hash(x)] for x in X]

    def inverse_transform(self, X):
        return [' '.join(x) for x in X]


class TokensEncoder(TransformerMixin, BaseEstimator):
    class PaddingToken:
        def __repr__(self):
            return "<PAD>"

    class UnknownToken:
        def __repr__(self):
            return "<UNK>"

    def __init__(self, max_len, max_vocab, min_freq):
        self.max_len = max_len
        self.max_vocab = max_vocab
        self.min_freq = min_freq

    def fit(self, X, y=None):
        token_freq = Counter()
        for sentence in X:
            token_freq.update(sentence)

        vocab = [token for token, count in token_freq.most_common(self.max_vocab)
                 if count >= self.min_freq]
        vocab.insert(0, self.PaddingToken())
        vocab.insert(1, self.UnknownToken())
        self.token2code = {token: i for i, token in enumerate(vocab)}
        self.vocab = vocab
        return self

    def transform(self, X):
        res = np.zeros((len(X), self.max_len), dtype=np.int)
        for i, sentence in enumerate(X):
            codes = [self.token2code.get(token, 1) for token in sentence[:self.max_len]]
            sentence_len = min(len(sentence), self.max_len)
            res[i, :sentence_len] = np.array(codes)
        return res

    def inverse_transform(self, X):
        return [[str(self.vocab[token_code]) for token_code in x if token_code != 0] for x in X]


class ToTensor(TransformerMixin, BaseEstimator):
    def __init__(self, dtype=None):
        self.dtype = dtype

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return torch.tensor(X, dtype=self.dtype)

    def inverse_transform(self, X):
        return X.numpy()


class MultiLabelEncoder(TransformerMixin, BaseEstimator):
    def fit(self, X, y=None):
        self.vocab = sorted(set([label for row in X for label in row]))
        return self

    def transform(self, X):
        return [[(_class in row) for _class in self.vocab] for row in X]


class To2DFloatArray(TransformerMixin, BaseEstimator):
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = np.array(X, dtype=np.object)
        X[X == ''] = np.nan
        return X.astype(np.float32).reshape(len(X), -1)

    def inverse_transform(self, X):
        return np.squeeze(X, axis=-1)


class Log(TransformerMixin, BaseEstimator):
    def __init__(self, auto_scale):
        self.auto_scale = auto_scale

    def fit(self, X, y=None):
        min_ = min(X)
        self.offset = 1 - min_ if (self.auto_scale and min_ < 1) else 0
        return self

    def transform(self, X):
        return np.log(X + self.offset)

    def inverse_transform(self, X):
        return np.exp(X) - self.offset



from functools import wraps
import torch

__all__ = [
    'mappify',
    'batch_to_device',
]


def mappify(func):
    @wraps(func)
    def inner(X, **kwargs):
        return [func(x, **kwargs) for x in X]
    return inner


def batch_to_device(batch, device):
    if isinstance(batch, torch.Tensor):
        return batch.to(device)
    elif isinstance(batch, (list, tuple)):
        res = [batch_to_device(x, device) for x in batch]
        return res if isinstance(batch, list) else tuple(res)
    elif isinstance(batch, dict):
        return {k: batch_to_device(v, device) for k, v in batch.items()}



import torch
from torch.optim import Adam
from torch.optim.lr_scheduler import OneCycleLR
from fastprogress.fastprogress import master_bar, progress_bar
from collections import defaultdict
import inspect
import matplotlib.pyplot as plt

from onfire.utils import batch_to_device

all = [
    'SupervisedRunner',
]


class SupervisedRunner:
    def __init__(self, model, loss_fn):
        self.model = model
        self.loss_fn = loss_fn

    def fit(self, train_dl, valid_dl, epochs, lr, metrics=None, optimizer=None, scheduler=None):
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model.to(device)
        optimizer = optimizer or Adam(self.model.parameters(), lr)
        if scheduler != False:
            scheduler = scheduler or OneCycleLR(optimizer, lr, epochs * len(train_dl))
        else:
            scheduler = None
        self.train_stats = TrainTracker(metrics, validate=(valid_dl is not None))
        bar = master_bar(range(epochs))
        bar.write(self.train_stats.metrics_names, table=True)

        for epoch in bar:
            self.model.train()
            for batch in progress_bar(train_dl, parent=bar):
                batch = batch_to_device(batch, device)
                loss = self._train_batch(batch, optimizer, scheduler)
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                if scheduler:
                    scheduler.step()
                self.train_stats.update_train_loss(loss)

            valid_outputs = []
            if valid_dl:
                self.model.eval()
                for batch in progress_bar(valid_dl, parent=bar):
                    batch = batch_to_device(batch, device)
                    output = self._valid_batch(batch)
                    valid_outputs.append(output)

            self.train_stats.log_epoch_results(valid_outputs)
            bar.write(self.train_stats.get_metrics_values(), table=True)

    def predict(self, dl, include_target=False):
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model.to(device)
        self.model.eval()
        preds, ys = [], []
        for batch in progress_bar(dl):
            batch = batch_to_device(batch, device)
            pred, y = self._predict_batch(batch, include_target)
            preds.append(pred)
            ys.append(y)
        preds = torch.cat(preds)
        return (preds, torch.cat(ys)) if include_target else preds

    def _train_batch(self, batch, optimizer, scheduler):
        *xb, yb = batch
        output = self.model(*xb)
        return self.loss_fn(output, yb)

    def _valid_batch(self, batch):
        *xb, yb = batch
        with torch.no_grad():
            output = self.model(*xb)
            loss = self.loss_fn(output, yb)
        return {'loss': loss.item(), 'y_true': yb.cpu(), 'y_pred': output.cpu()}

    def _predict_batch(self, batch, include_target):
        xb = batch[:-1] if len(batch) > 1 else [batch[0]]
        yb = batch[-1].cpu() if include_target and len(batch) > 1 else None
        with torch.no_grad():
            output = self.model(*xb)
        return output.cpu(), yb


class TrainTracker:
    def __init__(self, metrics, validate):
        if validate:
            self.valid_loss = []
            metrics = metrics if isinstance(metrics, (list, tuple)) else [metrics]
            self.metrics = [Metric(metric_fn) for metric_fn in metrics if metric_fn]
        self.train_smooth_loss = ExponentialMovingAverage()
        self.train_loss = []
        self.epoch = 0
        self.validate = validate

    @property
    def metrics_names(self):
        default_metrics = ['epoch', 'train_loss']
        metrics = []
        if self.validate:
            metrics.append('valid_loss')
            metrics.extend([metric.name for metric in self.metrics])
        return default_metrics + metrics

    def update_train_loss(self, loss):
        self.train_smooth_loss.update(loss.item())

    def log_epoch_results(self, valid_output):
        self.epoch = self.epoch + 1
        self.train_loss.append(self.train_smooth_loss.value)

        if self.validate:
            valid_output = self._process_valid_output(valid_output)
            valid_loss = valid_output['loss'].mean().item()
            for metric in self.metrics:
                metric.update(**valid_output)
            self.valid_loss.append(valid_loss)

    def get_metrics_values(self, decimals=5):
        default_metrics = [self.epoch, self.train_loss[-1]]
        metrics = []
        if self.validate:
            metrics.append(self.valid_loss[-1])
            metrics.extend([metric.value for metric in self.metrics])
        res = default_metrics + metrics
        return [str(x) if isinstance(x, int) else str(round(x, decimals)) for x in res]

    def _process_valid_output(self, valid_output):
        res = defaultdict(list)
        for d in valid_output:
            for k, v in d.items():
                v = v if isinstance(v, torch.Tensor) else torch.tensor(v)
                v = v if len(v.shape) else v.view(1)
                res[k].append(v)
        return {k: torch.cat(v) for k, v in res.items()}

    def plot_loss(self):
        fig, ax = plt.subplots()
        ax.plot(self.train_loss, label='train')
        ax.plot(self.valid_loss, label='valid')
        ax.legend()


class ExponentialMovingAverage():
    def __init__(self, beta=0.1):
        self.beta = beta
        self.initialized = False

    def update(self, value):
        if self.initialized:
            self.mean = value * self.beta + self.mean * (1 - self.beta)
        else:
            self.mean = value
            self.initialized = True

    @property
    def value(self):
        return self.mean


class Metric:
    def __init__(self, metric_fn):
        self.metric_fn = metric_fn
        self.name = metric_fn.__name__ if inspect.isfunction(metric_fn) else str(metric_fn)
        self.value = None

    def update(self, **kwargs):
        y_true, y_pred = kwargs['y_true'], kwargs['y_pred']
        self.value = self.metric_fn(y_true, y_pred)



import torch
from collections import OrderedDict
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.impute import SimpleImputer
from abc import ABC, abstractmethod

from .transformers import (
    Projector, LabelEncoder, BasicTokenizer, TokensEncoder,
    ToTensor, MultiLabelEncoder, To2DFloatArray, Log)
from .embedders import ConcatEmbeddings, PassThrough, MeanOfEmbeddings

__all__ = [
    'CategoricalFeature',
    'TextFeature',
    'ContinuousFeature',
    'FeatureGroup',
    'SingleLabelTarget',
    'MultiLabelTarget',
    'ContinuousTarget',
]


class BaseField(ABC, TransformerMixin, BaseEstimator):
    def __init__(self, key, preprocessor, custom_tfms=None, dtype=None):
        tfms = []
        if key: tfms.append(Projector(key))
        if preprocessor: tfms.append(preprocessor)
        if custom_tfms: tfms.extend(custom_tfms)
        tfms.append(ToTensor(dtype=dtype))
        self.pipe = make_pipeline(*tfms)

    def transform(self, X):
        return self.pipe.transform(X)

    def inverse_transform(self, X):
        return self.pipe.inverse_transform(X)

    @property
    @abstractmethod
    def output_dim(self):
        pass


class BaseFeature(BaseField):
    @abstractmethod
    def build_embedder(self):
        pass

    @property
    def embedder(self):
        pass


class CategoricalFeature(BaseFeature):
    def __init__(self, key=None, preprocessor=None, emb_dim=None):
        self.key = key
        self.preprocessor = preprocessor
        self.emb_dim = emb_dim
        self.categorical_encoder = LabelEncoder()

        tfms = [self.categorical_encoder]
        super().__init__(self.key, self.preprocessor, tfms, dtype=torch.int64)

    def fit(self, X, y=None):
        self.pipe.fit(X)
        self.vocab = self.categorical_encoder.vocab
        self.emb_dim = self.emb_dim or min(len(self.vocab) // 2, 50)
        return self

    def build_embedder(self):
        self._embedder = torch.nn.Embedding(len(self.vocab), self.emb_dim)
        return self.embedder

    @property
    def output_dim(self):
        return self.emb_dim

    @property
    def embedder(self):
        return self._embedder


class TextFeature(BaseFeature):
    def __init__(self, key=None, preprocessor=None, max_len=50, max_vocab=50000,
                 min_freq=3, emb_dim=100, tokenizer=None, embedder_cls=None,
                 embedder_vocab_size_param='vocab_size', embedder_args=None):
        self.max_len = max_len
        self.max_vocab = max_vocab
        self.key = key
        self.preprocessor = preprocessor
        self.tokenizer = tokenizer or BasicTokenizer()
        self.min_freq = min_freq
        self.emb_dim = emb_dim
        self.token_encoder = TokensEncoder(self.max_len, self.max_vocab, self.min_freq)
        self.embedder_cls = embedder_cls or MeanOfEmbeddings
        self.embedder_vocab_size_param = embedder_vocab_size_param
        self.embedder_args = embedder_args or {}
        if self.embedder_cls == MeanOfEmbeddings:
            self.embedder_args['emb_dim'] = self.emb_dim

        tfms = [self.tokenizer, self.token_encoder]
        super().__init__(self.key, self.preprocessor, tfms, dtype=torch.int64)

    def fit(self, X, y=None):
        self.pipe.fit(X)
        self.vocab = self.token_encoder.vocab
        self.embedder_args[self.embedder_vocab_size_param] = len(self.vocab)
        return self

    def build_embedder(self):
        self._embedder = self.embedder_cls(**self.embedder_args)
        sample_input = torch.randint(len(self.vocab), (2, self.max_len))
        self.emb_dim = self.embedder(sample_input).shape[1]
        return self.embedder

    @property
    def output_dim(self):
        return self.emb_dim

    @property
    def embedder(self):
        return self._embedder


class ContinuousFeature(BaseFeature):
    def __init__(self, key=None, preprocessor=None, imputer=None, scaler=None, log=False,
                 log_auto_scale=True):
        self.key = key
        self.preprocessor = preprocessor
        self.imputer = (imputer or SimpleImputer()) if imputer != False else None
        self.scaler = (scaler or StandardScaler()) if scaler != False else None
        self.log = log
        self.log_auto_scale = log_auto_scale

        tfms = []
        tfms.append(To2DFloatArray())
        if self.imputer: tfms.append(self.imputer)
        if self.log: tfms.append(Log(auto_scale=self.log_auto_scale))
        if self.scaler: tfms.append(self.scaler)
        super().__init__(self.key, self.preprocessor, tfms, dtype=torch.float32)

    def fit(self, X, y=None):
        self.pipe.fit(X)
        self.emb_dim = self.transform([X[0]]).shape[1]
        return self

    def build_embedder(self):
        self._embedder = PassThrough()
        return self.embedder

    @property
    def output_dim(self):
        return self.emb_dim

    @property
    def embedder(self):
        return self._embedder


class FeatureGroup(BaseFeature):
    def __init__(self, fields):
        self.fields = OrderedDict(fields)

    def fit(self, X, y=None):
        for field in self.fields.values():
            field.fit(X)
        return self

    def transform(self, X, y=None):
        return [field.transform(X) for field in self.fields.values()]

    def inverse_transform(self, X):
        tmp = [field.inverse_transform(X[i]) for i, field in enumerate(self.fields.values())]
        res = []
        for i in range(len(X[0])):
            d = {}
            for field in tmp:
                d.update(field[i])
            res.append(d)
        return res

    def build_embedder(self):
        self._embedder = ConcatEmbeddings(self.fields)
        return self.embedder

    @property
    def output_dim(self):
        return self.embedder.output_dim

    @property
    def embedder(self):
        return self._embedder


class SingleLabelTarget(BaseField):
    def __init__(self, key=None, preprocessor=None, dtype=torch.int64):
        self.key = key
        self.preprocessor = preprocessor
        self.categorical_encoder = LabelEncoder(is_target=True)
        self.dtype = dtype

        tfms = [self.categorical_encoder]
        super().__init__(self.key, self.preprocessor, tfms, dtype=self.dtype)

    def fit(self, X, y=None):
        self.pipe.fit(X)
        self.classes = self.categorical_encoder.vocab
        return self

    @property
    def output_dim(self):
        return len(self.classes)


class MultiLabelTarget(BaseField):
    def __init__(self, key=None, preprocessor=None, dtype=torch.float32):
        self.key = key
        self.preprocessor = preprocessor
        self.multi_label_encoder = MultiLabelEncoder()
        self.dtype = dtype

        tfms = [self.multi_label_encoder]
        super().__init__(self.key, self.preprocessor, tfms, dtype=self.dtype)

    def fit(self, X, y=None):
        self.pipe.fit(X)
        self.classes = self.multi_label_encoder.vocab
        return self

    @property
    def output_dim(self):
        return len(self.classes)


class ContinuousTarget(BaseField):
    def __init__(self, key=None, preprocessor=None, log=False, log_auto_scale=False):
        self.key = key
        self.preprocessor = preprocessor
        self.log = log
        self.log_auto_scale = log_auto_scale

        tfms = []
        tfms.append(To2DFloatArray())
        if self.log: tfms.append(Log(auto_scale=self.log_auto_scale))

        super().__init__(self.key, self.preprocessor, tfms, dtype=torch.float32)

    def fit(self, X, y=None):
        self.pipe.fit(X)
        self.out_dim = self.transform([X[0]]).shape[1]
        return self

    @property
    def output_dim(self):
        return self.out_dim



from setuptools import setup
from setuptools.extern import packaging
import ast

with open('onfire/__init__.py') as f:
    for line in f:
        if line.startswith('__version__'):
            version = ast.parse(line).body[0].value.s
            version = str(packaging.version.Version(version))
            break

with open('README.md') as f:
    long_description = f.read()

setup(
    name='pytorch-onfire',
    version=version,
    author='Jose Fernandez Portal, Rafael Carrascosa',
    author_email='jose.fp@gmail.com',
    description='PyTorch meets Sklearn Pipelines.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/joshfp/pytorch-onfire',
    license='MIT',
    packages=[
        'onfire',
        'onfire.colab',
    ],
    install_requires=[
        'torch',
        'scikit-learn',
        'Unidecode',
        'lmdb',
        'msgpack',
        'fastprogress',
        'matplotlib',
    ],
    python_requires='>=3.6',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
